#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import os
import argparse
import numpy as np
import joblib
from collections import defaultdict, OrderedDict, Counter

from .sirms_file_class import SirmsFile, mol_frag_sep
from .predict import load_multi_obj


def load_sirms(fname):
    with open(fname) as f:
        descr_names = f.readline().strip().split("\t")[1:]
        case_names = []
        x = []
        for line in f:
            tmp = line.strip().split("\t")
            case_names.append(tmp[0])
            x.append(tuple(map(float, tmp[1:])))
    return descr_names, case_names, np.asarray(x)


def split_mol_frag_names(names, ids):
    """
    names: list of names in format mol_name#frag_name
    ids: list of boolean
    """
    mol_names = []
    frag_names = []
    for i, el in enumerate(ids):
        if el:
            tmp = names[i].split(mol_frag_sep)
            mol_names.append(tmp[0])
            frag_names.append(tmp[1])
    return mol_names, frag_names


def prepare_dataset(x_train, x_frag, prop_name, descr_names, x_train_mol_names, x_frag_mol_names):
    """
    x_train: numpy array with descriptors for training set
    x_frag: numpy array with descriptors for dataset after removal of specified fragments
    prop_name: name of atomic property, which contribution is estimated
    descr_names: list of descriptors names in x_train and x_frag
    x_train_mol_names: list of molecule names in x_train
    x_frag_mol_names: list of molecule names in x_frag

    http://stackoverflow.com/questions/26086645/manipulation-with-arrays-in-numpy
    """
    ids = np.array(["|" + prop_name + "|" in el for el in descr_names])
    rows = np.where(np.array(x_train_mol_names) == np.array(x_frag_mol_names)[:, np.newaxis])[1]
    output = np.where(ids, x_frag, x_train[rows])
    return output


def save_contrib(fname, contrib_dict, frag_full_names, long_format, save_frag_ids):
    if save_frag_ids:
        frag_names = frag_full_names
    else:
        frag_names = [v.rsplit('#', 1)[0] for v in frag_full_names if v.find(mol_frag_sep) > -1]

    if not long_format:
        with open(fname, "wt") as f:
            f.write("Model_Contribution\t" + "\t".join(frag_names) + "\n")
            for model_name in contrib_dict.keys():
                for prop_name in contrib_dict[model_name].keys():
                    f.write(model_name + "_" + prop_name + "\t" +
                            "\t".join(["{0:.6f}".format(i) for i in contrib_dict[model_name][prop_name]]) + "\n")
    else:
        with open(fname, "wt") as f:
            if save_frag_ids:
                f.write("Compound\tFragment\tFrag_id\tModel\tContribution_type\tContribution_value\n")
            else:
                f.write("Compound\tFragment\tModel\tContribution_type\tContribution_value\n")
            for model_name in contrib_dict.keys():
                for prop_name in contrib_dict[model_name].keys():
                    for frag_full_name, value in zip(frag_names, contrib_dict[model_name][prop_name]):
                        if save_frag_ids and frag_full_name.find(mol_frag_sep) > -1:
                            mol_name, frag_name = frag_full_name.split(mol_frag_sep)
                            frag_name, frag_id = frag_name.rsplit('#', 1)
                            f.write('\t'.join((mol_name, frag_name, frag_id, model_name, prop_name, "{0:.6f}".format(value))) + '\n')
                        else:
                            f.write('\t'.join(frag_full_name.split(mol_frag_sep)) + '\t' + model_name + '\t' + prop_name + '\t' + "{0:.6f}".format(value) + '\n')


def adjust_dataset(x, x_col_names, ref_col_names, default_value=0):

    # init output with default value
    if default_value != 0:
        output = np.empty((np.shape(x)[0], len(ref_col_names)))
        output.fill(default_value)
    else:
        output = np.zeros((np.shape(x)[0], len(ref_col_names)))

    # get indices in output (-1 means no such column in output)
    ids = [-1] * len(x_col_names)
    for i, col in enumerate(x_col_names):
        try:
            ids[i] = ref_col_names.index(col)
        except ValueError:
            pass

    keep_cols = [i != -1 for i in ids]
    ids = [i for i in ids if i != -1]

    output[::, np.array(ids)] = x[::, np.array(keep_cols)]

    return output


def predict(x, model, model_name, model_type):
   
    pred = None

    if model_type == 'reg':
        pred = model.predict(x)
        if model_name == 'pls':
            pred = pred.reshape(pred.shape[0] * pred.shape[1])
        pred = pred.tolist()

    elif model_type == 'class':
        pos_id = model.classes_.tolist().index(1)
        pred = [i[pos_id] for i in model.predict_proba(x)]
    return pred


def predict_avg(x, models, model_name, model_type):
    """
    Make prediction by all models and calculate average prediction
    :param x: numpy array of X
    :param models: list of models
    :param model_name: name of model (svm, rf, ...)
    :param model_type: 'class' or 'reg'
    :return:
    """
    pred = []
    for m in models:
        pred.append(predict(x, m, model_name, model_type))
    return np.array(pred).mean(0)


def read_mol_names_from_activity_file(fname):
    with open(fname) as f:
        f.readline()
        output = [line.split()[0] for line in f]
    return output


def adjust_mols(keep_mol_names, mol_names, x):
    # the len of mol_names should be equal to number of rows of x
    ids = []
    output_mol_names = []
    # count mol names to remove mols withput fragments 
    c = Counter((mol_name.split(mol_frag_sep)[0] for mol_name in mol_names))
    for i, mol_name in enumerate(mol_names):
        m_name = mol_name.split(mol_frag_sep)[0]
        if m_name in keep_mol_names and c[m_name] > 1:
            ids.append(i)
            output_mol_names.append(mol_name)

    return output_mol_names, x[ids]


def main_params(x_fname, out_fname, model_names, model_dir, prop_names, model_type, verbose, save_pred, input_format,
                long_format, save_frag_ids, activity_file):

    if save_pred:
        save_pred_fname = os.path.splitext(out_fname)[0] + "_pred.txt"
        if os.path.isfile(save_pred_fname):
            print("File with predictions was erased.")
            os.remove(save_pred_fname)

    sirms_file = SirmsFile(x_fname, file_format=input_format, frag_file=True, chunks=1000)

    # scale
    if os.path.isfile(os.path.join(model_dir, "scale.pkl")):
        scale = joblib.load(os.path.join(model_dir, "scale.pkl"))
    else:
        scale = None

    # frag_contrib = dict()
    frag_contrib = OrderedDict()

    # for adjusting data sets before prediction (solving different order of descriptors)
    ref_var_names = joblib.load(os.path.join(model_dir, "var_names.pkl"))

    if activity_file:
        used_mol_names = set(read_mol_names_from_activity_file(activity_file))
    if model_type == 'reg' or model_type == 'class':

        for model_name in model_names:

            if verbose:
                print("Contributions calculation. %s model proceed" % model_name)

            frag_contrib[model_name] = defaultdict(list)

            frag_full_names = []  # used for save results - reinit for each model but should be identical

            sirms_file.reset_read()
            mol_names, var_names, x = sirms_file.read_next()

            while mol_names:

                if activity_file:
                    mol_names, x = adjust_mols(used_mol_names, mol_names, x)
                    
                if mol_names:   # non-empty list after removal of mols
                    x = adjust_dataset(x, var_names, ref_var_names)
                    var_names = ref_var_names[:]

                    if scale is not None:
                        x = scale.transform(x)

                    x_frag_ids = [mol_frag_sep in n for n in mol_names]
                    x_train_ids = [not i for i in x_frag_ids]

                    frag_full_names.extend((mol_name for i, mol_name in zip(x_frag_ids, mol_names) if i))

                    x_frag_mol_names, x_frag_frag_names = split_mol_frag_names(mol_names, x_frag_ids)
                    x_train_mol_names = [mol_names[i] for i, el in enumerate(x_train_ids) if el]

                    models = load_multi_obj(os.path.join(model_dir, model_name + ".pkl"))
                    train_pred = predict_avg(x[np.asarray(x_train_ids), ], models, model_name, model_type)
                    train_pred = dict(zip(x_train_mol_names, train_pred))

                    for prop_name in prop_names:
                        models = load_multi_obj(os.path.join(model_dir, model_name + ".pkl"))
                        if prop_name == "overall":
                            frag_pred = predict_avg(x[np.asarray(x_frag_ids), ], models, model_name, model_type)
                        else:
                            x_frag_prep = prepare_dataset(x[np.asarray(x_train_ids), ],
                                                          x[np.asarray(x_frag_ids), ],
                                                          prop_name,
                                                          var_names,
                                                          x_train_mol_names,
                                                          x_frag_mol_names)
                            frag_pred = predict_avg(x_frag_prep, models, model_name, model_type)

                        try:
                            frag_contrib[model_name][prop_name].extend([train_pred[mol_name] - frag_pred[i] for i, mol_name in enumerate(x_frag_mol_names)])
                        except KeyError:
                            pass

                        if save_pred:
                            with open(save_pred_fname, "at") as f:
                                for k, v in train_pred.items():
                                    f.write(k + "\t" + model_name + "\t" + prop_name + "\t" + str(v) + "\n")
                                for m_name, f_name, pred_value in zip(x_frag_mol_names, x_frag_frag_names, frag_pred):
                                    f.write(m_name + mol_frag_sep + f_name + "\t" + model_name + "\t" + prop_name + "\t" + str(pred_value) + "\n")

                mol_names, var_names, x = sirms_file.read_next()

    save_contrib(out_fname, frag_contrib, frag_full_names, long_format, save_frag_ids)


def main():

    parser = argparse.ArgumentParser(description='Calculate contributions of molecular fragments based on QSAR models.')
    parser.add_argument('-i', '--input', metavar='frag_descriptors.txt', required=True,
                        help='input text file with descriptors for compounds with removed fragments.')
    parser.add_argument('-f', '--input_format', metavar='txt|svm', default='txt',
                        help='format of input file with descriptors for compounds with removed fragments (txt|svm). '
                             'Svm sparse format file ha to have two files with the same name and extensions '
                             'rownames and colnames. Default: txt.')
    parser.add_argument('-o', '--output', metavar='contributions.txt', required=True,
                        help='output text file with fragments contributions.')
    parser.add_argument('--output_long_format', action='store_true', default=False,
                        help='store output in long instead of short format. Long format is easier to parse but takes '
                             'more disk space. Default: false.')
    parser.add_argument('-a', '--activity_file', metavar='activity_file.txt', required=False, default=None,
                        help='path to the file with activity values. If specified only molecules mention in this file '
                             'will be used for calculation of contributions. Default: None.')
    parser.add_argument('-m', '--models', metavar='[rf gbm svr pls knn]', required=True, nargs='*',
                        help='file name of saved models (without extension)')
    parser.add_argument('-d', '--models_dir', metavar='path_to_models', required=True,
                        help='path to stored model files')
    parser.add_argument('-p', '--properties', metavar='[overall CHARGE HB ...]', nargs='*', default=["overall"],
                        help='name of atomic properties which should be considered during calculation of '
                             'contribution of fragments')
    parser.add_argument('-t', '--model_type', metavar='reg|class', default='reg',
                        help='models type: reg for regression and class for classification')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='show progress on the screen.')
    parser.add_argument('-s', '--save_intermediate_prediction', action='store_true', default=False,
                        help='save intermediate prediction to text file.')
    parser.add_argument('--save_frag_ids', action='store_true', default=False,
                        help='save frag id in each molecule. For wide format output (default) it will return '
                             'frag_name#frag_id. For long format output it will return additional column named '
                             'Frag_ID.')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "input": x_fname = v
        if o == "output": out_fname = v
        if o == "models": model_names = v
        if o == "models_dir": model_dir = v
        if o == "properties": prop_names = v
        if o == "model_type": model_type = v
        if o == "verbose": verbose = v
        if o == "save_intermediate_prediction": save_pred = v
        if o == "input_format": input_format = v
        if o == "output_long_format": long_format = v
        if o == "save_frag_ids": save_frag_ids = v
        if o == "activity_file": activity_file = v
    if input_format not in ['txt', 'svm']:
        print("Wrong input file format - %s. Only txt and svm file formats are allowed." % input_format)
        exit()

    main_params(x_fname, out_fname, model_names, model_dir, prop_names, model_type, verbose, save_pred,
                input_format, long_format, save_frag_ids, activity_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 08-06-2017
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2017
# license         : LGPLv3
#==============================================================================

import argparse
from rdkit import Chem


def main_params(input_sdf_fname, field_name, output_sdf_fname):

    w = Chem.SDWriter(output_sdf_fname)
    w.SetKekulize(False)
    i = 1

    for m in Chem.SDMolSupplier(input_sdf_fname, False, False, False):
        if m:
            if field_name and m.HasProp(field_name):
                m.SetProp('_Name', m.GetProp(field_name))
            else:
                m.SetProp('_Name', 'MolID_%i' % i)
                i += 1
            w.write(m)

    w.close()


def main():

    parser = argparse.ArgumentParser(description='Insert mol titles to sdf file.')
    parser.add_argument('-i', '--input', metavar='input.sdf', required=True,
                        help='input sdf file.')
    parser.add_argument('-o', '--output', metavar='output.sdf', required=True,
                        help='output sdf file.')
    parser.add_argument('-f', '--field_name', default=None,
                        help='field name in sdf file which will be copied to molecule title, '
                             'if omitted compounds will ne enumerated sequentially with MolID_ prefix.')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "input": input_fname = v
        if o == "output": output_fname = v
        if o == "field_name": field_name = v

    main_params(input_fname, field_name, output_fname)


if __name__ == '__main__':
    main()


#!/usr/bin/env python3
# author          : Pavel
# date            : 04.12.15
# version         : 0.1
# python_version  : 3
# copyright       : Pavel 2015
# license         : LGPLv3
#==============================================================================

import os
import re
import argparse

from datetime import datetime
from rdkit import Chem
from rdkit.Chem import AllChem, rdMMPA

from .mol_context import get_canon_context_core


patt = re.compile("\[\*\:[0-9]+\]")  # to change CC([*:1])O to CC([*])O


def replace_no2(mol):
    query = Chem.MolFromSmarts('n(:o):o')
    repl = Chem.MolFromSmiles('[N+](=O)[O-]')
    return AllChem.ReplaceSubstructs(mol, query, repl, replaceAll=True)[0]


def fragment_mol(mol, query, max_cuts, keep_stereo, radius, max_size):
    # returns list of lists: [['F', [0]], ['C#N', [3, 4]], ... ]

    def get_atom_prop(molecule, prop="Index", only_heavy=True):
        res = []
        for a in molecule.GetAtoms():
            if only_heavy and a.GetAtomicNum() > 1:
                try:
                    res.append(a.GetIntProp(prop))
                except KeyError:
                    continue
        return tuple(sorted(res))

    def get_frag_name(context, core, radius, keep_stereo):
        line = []
        for r in radius:
            env_smi, core_smi = get_canon_context_core(context, core, r, keep_stereo)
            if r == 0:  # for radius = 0 there is no env (empty string)
                line.append(core_smi)
            else:
                if env_smi and core_smi:
                    line.append('%s|%s' % (core_smi, env_smi))
        return '||'.join(line) if line else None

    # modify representation of NO2 groups to charged version
    mol = replace_no2(mol)
    err = Chem.SanitizeMol(mol, catchErrors=True)
    if err != 0:
        print('Molecule %s failed to sanitize due to: ' % mol.GetProp("_Name") + str(err))
        return []

    output = []

    for atom in mol.GetAtoms():
        atom.SetIntProp("Index", atom.GetIdx())

    frags = rdMMPA.FragmentMol(mol, pattern=query, maxCuts=max_cuts, resultsAsMols=True, maxCutBonds=30)

    for core, chains in frags:
        if core is None:  # single cut
            components = list(Chem.GetMolFrags(chains, asMols=True))
            ids_0 = get_atom_prop(components[0])
            ids_1 = get_atom_prop(components[1])
            if Chem.MolToSmiles(components[0]) != '[H][*:1]':  # context cannot be H
                if max_size is None or components[1].GetNumHeavyAtoms() <= max_size:
                    frag_name = get_frag_name(components[0], components[1], radius, keep_stereo)
                    if frag_name:
                        output.append((frag_name, ids_1))
            if Chem.MolToSmiles(components[1]) != '[H][*:1]':  # context cannot be H
                if max_size is None or components[0].GetNumHeavyAtoms() <= max_size:
                    frag_name = get_frag_name(components[1], components[0], radius, keep_stereo)
                    if frag_name:
                        output.append((frag_name, ids_0))
        else:   # multiple cuts
            if Chem.MolToSmiles(chains).find('[H][*:1]') == -1:  # context cannot be H
                if max_size is None or core.GetNumHeavyAtoms() <= max_size:
                    frag_name = get_frag_name(chains, core, radius, keep_stereo)
                    if frag_name:
                        output.append((frag_name, get_atom_prop(core)))

    return output


def main_params(in_sdf, out_txt, query, max_cuts, keep_stereo, radius, verbose, error_fname, max_size=None):

    with open(out_txt, 'wt') as f:
        for i, mol in enumerate(Chem.SDMolSupplier(in_sdf, sanitize=False, removeHs=False)):
            if mol is not None:
                if verbose:
                    print(mol.GetProp("_Name") + ' is processing')
                res = fragment_mol(mol, query, max_cuts, keep_stereo, radius, max_size)
                for item in res:
                    ids = [i + 1 for i in item[1]]  # save as 1-based ids
                    f.write(mol.GetProp("_Name") + '\t' + item[0] + '\t' + '\t'.join(map(str, ids)) + '\n')
            else:
                print('Molecule #%i was skipped due to error' % i)
                with open(error_fname, 'at') as f_err:
                    f_err.write('%s\t%s\tMolecule #%i\n' % (datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                                                            os.path.basename(__file__), i + 1))


def main():

    parser = argparse.ArgumentParser(description='Generate all possible fragments and find indices of them in '
                                                 'molecules from input sdf-file.')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=True,
                        help='input sdf file with standardized structures, molecules should have titles.')
    parser.add_argument('-o', '--out', metavar='output.txt', required=True,
                        help='output text file; each line contains tab-separated fields: name of a molecule, '
                             'name of a fragment and list of corresponding atom numbers.')
    parser.add_argument('-q', '--query', metavar='smarts', default='[#6+0;!$(*=,#[!#6])]!@!=!#[!#1]',
                        help='SMARTS string to match bonds to cleave. Default: [#6+0;!$(*=,#[!#6])]!@!=!#[!#1]')
    parser.add_argument('-u', '--upper_cuts_number', metavar='integer', default=3,
                        help='maximal number of bonds cleaved simultaneously. Default: 3')
    parser.add_argument('-r', '--radius', metavar='integer', default=[0], nargs='*',
                        help='radius of molecular context (in bonds) which will be taken into account. '
                             '0 means no context. Several values separated by space can be specified. '
                             'The output fragment names will consist of core_smi_1|env_smi_1||core_smi_2|env_smi_2.'
                             ' Default: [0].')
    parser.add_argument('--max_size', metavar='integer', default=None, required=False, type=int,
                        help='maximum number of heavy atoms in output fragments. Default: no restriction.')
    parser.add_argument('-s', '--keep_stereo', action='store_true', default=False,
                        help='set this flag to keep stereo in context and core parts.')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='show progress on the screen.')
    parser.add_argument('-e', '--error_file', metavar='log_file_name.txt', default="rdkit_errors.txt",
                        help='save a number of molecules which cause error. Default file name: rdkit_errors.txt.')


    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "in": in_sdf = v
        if o == "out": out_txt = v
        if o == "query": query = v
        if o == "upper_cuts_number": max_cuts = int(v)
        if o == "verbose": verbose = v
        if o == "error_file": error_fname = v
        if o == "keep_stereo": keep_stereo = v
        if o == "radius": radius = tuple(sorted(map(int, v)))
        if o == "max_size": max_size = v

    main_params(in_sdf, out_txt, query, max_cuts, keep_stereo, radius, verbose, error_fname, max_size)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# author          : Pavel
# date            : 05.11.15
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel 2015
# license         : LGPLv3
#==============================================================================

import os
import argparse
import numpy as np
import joblib
from collections import defaultdict, Counter
import re
import math
import operator

from .sirms_file_class import SirmsFile


def get_var_names(fname, fformat):
    # fformat = txt or svm
    output = None
    if fformat == 'txt':
        output = open(fname).readline().strip().split('\t')[1:]
    elif fformat == 'svm':
        with open(os.path.splitext(fname)[0] + '.colnames') as f:
            output = []
            for line in f:
                output.append(line.strip())
    return output


def adjust_dataset(x, x_col_names, ref_col_names, default_value=0):

    # init output with default value
    if default_value != 0:
        output = np.empty((np.shape(x)[0], len(ref_col_names)))
        output.fill(default_value)
    else:
        output = np.zeros((np.shape(x)[0], len(ref_col_names)))

    # get indices in output (-1 means no such column in output)
    ids = [-1] * len(x_col_names)
    for i, col in enumerate(x_col_names):
        try:
            ids[i] = ref_col_names.index(col)
        except ValueError:
            pass

    keep_cols = [i != -1 for i in ids]
    ids = [i for i in ids if i != -1]

    output[::, np.array(ids)] = x[::, np.array(keep_cols)]

    return output


def predict(x, model, model_name, model_type):

    pred = None

    if model_type == 'reg':
        pred = model.predict(x)
        if re.match("pls^|pls_", model_name):  # we added _0,1,2..
            pred = pred.reshape(pred.shape[0] * pred.shape[1])
        pred = pred.tolist()

    elif model_type == 'class':
        pos_id = model.classes_.tolist().index(1)
        pred = [i[pos_id] for i in model.predict_proba(x)]

    return pred


def add_consensus(pred_array, pred_model_names, ids=None, current_model_name=None):
    # ids and current_model_name are used when we compute not overall consensus
    # but consensus of multiple models of single ML
    # method (svm_0. svm_1...)(example: ids=[0,1,2], current_model_name="svm")

    if np.shape(pred_array)[1] > 1:  # otherwise nothing to compute
        if not ids:  # all columns will be selected
            ids = range(np.shape(pred_array)[1])
        pred_array = np.append(pred_array, np.mean(pred_array[:, ids], 1).reshape((-1, 1)), 1)
        if current_model_name:  # case when we compute consensus for single ml model
            pred_model_names.append(current_model_name + "_" + "consensus")
        else:  # case when we compute overall consensus
            pred_model_names.append("consensus")
    return pred_array  # this arrary returned because it is not changed outside function.  pred_model_names ont returned they are changed


def save_predictions(fname, pred_array, pred_model_names, mol_names, ad_dict):
    # add mol_names ass the first column
    pred_array = np.column_stack((mol_names, pred_array))
    pred_model_names.insert(0, 'Compounds')
    # add AD values as additional columns
    if ad_dict is not None:
        for k, v in ad_dict.items():
            pred_array = np.column_stack((pred_array, v))
            pred_model_names.append(k)
    np.savetxt(fname, pred_array, '%s', delimiter='\t', header='\t'.join(pred_model_names))


def load_bound_box_constrains(fname):
    d = joblib.load(fname)
    return d['min_values'], d['max_values']


def check_bound_box(x, bound_box_constrains):
    # bound_box_constrains: tuple of tuples of min and max values
    a = np.greater_equal(x, bound_box_constrains[0])
    b = np.less_equal(x, bound_box_constrains[1])
    c = np.logical_and(a, b)
    d = np.all(c, axis=1)
    return d


def load_models(model_dir, model_names):
    """
    Read all model objects from each model file and store in the list.
    Returns corresponding names of models by adding _N (e.g. svm_0, etc)
    :param model_dir: path to dir where models are stored
    :param model_names: list of model names (file names without .pkl extension)
    :return: list of models and list of model names
    """
    models = []
    pred_model_names = []
    for m in model_names:
        for i, j in enumerate(load_multi_obj(os.path.join(model_dir, m + ".pkl"))):
            models.append(j)
            pred_model_names.append(m + "_" + str(i))
    return models, pred_model_names


def load_multi_obj(filename):
    with open(filename, "rb") as f:
        while True:
            try:
                yield joblib.load(f)
            except EOFError:
                break


def get_major_vote(data):
    """
    :param data: iterable
    :return: item which is most often occur
    """
    data = [i for i in data if not math.isnan(i)]
    return max(Counter(data).items(), key=operator.itemgetter(1))[0]


def main_params(x_fname, input_format, out_fname, model_names, model_dir, model_type, ad, verbose):

    if ad is not None:
        ad_dict = {item: [] for item in ad}
        bound_box_constrains = load_bound_box_constrains(os.path.join(model_dir, "bound_box.pkl"))
    else:
        ad_dict = None

    # scale
    if os.path.isfile(os.path.join(model_dir, "scale.pkl")):
        scale = joblib.load(os.path.join(model_dir, "scale.pkl"))
    else:
        scale = None

    pred = defaultdict(list)

    # load models
    models, pred_model_names = load_models(model_dir, model_names)

    input_sirms_file = SirmsFile(x_fname, file_format=input_format, chunks=1000)
    ref_var_names = joblib.load(os.path.join(model_dir, "var_names.pkl"))

    input_sirms_file.reset_read()
    mol_names, var_names, x = input_sirms_file.read_next()

    while mol_names:

        x = adjust_dataset(x, var_names, ref_var_names)

        if ad is not None:
            for ad_name in ad:
                ad_dict[ad_name].extend(check_bound_box(x, bound_box_constrains))

        if scale is not None:
            x = scale.transform(x)

        for model_name, m in zip(pred_model_names, models):
            pred[model_name].extend(predict(x, m, model_name, model_type))

        mol_names, var_names, x = input_sirms_file.read_next()

    # convert to numpy array

    pred = np.array([pred[k] for k in pred_model_names]).transpose()

    # add consensuses of multiple models if we have them (svm_0, ...1, ...2..etc) and final consensus
    if len(model_names) < len(pred_model_names):  # multiple models
        for model_name in model_names:
            if model_type == "class":
                cols = [i for i, j in enumerate(pred_model_names) if re.fullmatch(model_name + "_\d*", j)]
                pred = np.append(pred, np.mean(np.around(pred[:, cols]), 1).reshape((-1, 1)), 1)
                pred_model_names.append(model_name + "_consensus")
        # add final consensus
        if model_type == "class":
            cols = [i for i, j in enumerate(pred_model_names) if j.endswith("_consensus")]
            pred = np.append(pred, np.mean(np.around(pred[:, cols]), 1).reshape((-1, 1)), 1)
            pred_model_names.append("consensus")
    else:
        if model_type == "class":
            pred = np.append(pred, np.mean(np.around(pred), 1).reshape((-1, 1)), 1)
            pred_model_names.append("consensus")
        elif model_type == "reg":
            pred = np.append(pred, np.mean(pred, 1).reshape((-1, 1)), 1)
            pred_model_names.append("consensus")
    pred = np.around(pred, 3)
    save_predictions(out_fname, pred, pred_model_names, input_sirms_file.get_mol_names(), ad_dict)


def main():

    parser = argparse.ArgumentParser(description='Predict properties of compounds based on QSAR models.')
    parser.add_argument('-i', '--input', metavar='descriptors.txt', required=True,
                        help='input text file with descriptors for compounds with removed fragments.')
    parser.add_argument('-f', '--input_format', metavar='txt|svm', required=False, default='txt',
                        help='format of the input file with descriptors (txt|svm). Default: txt.')
    parser.add_argument('-o', '--output', metavar='predictions.txt', required=True,
                        help='output text file with predicitions. Contains predictions of all models and their consensus. Special case:if' ' there are multiple models of individual'
                             'ML methods in the models_dir, then consensus for each metod and consensus of those consensuses (final consensus) will be computed.')
    parser.add_argument('-m', '--models', metavar='[rf gbm svr pls knn]', required=True, nargs='*',
                        help='file names of saved models (without extension).')
    parser.add_argument('-d', '--models_dir', metavar='path_to_models', required=True,
                        help='path to stored model files. The file scale.pkl must be in that dir also.')
    parser.add_argument('-t', '--model_type', metavar='reg|class', default='reg',
                        help='models type: reg for regression and class for classification')
    parser.add_argument('-a', '--applicability_domain', metavar='none|bound_box', required=False, nargs='*', default=None,
                        help='name(s) of applicability domain(s) to apply. If several - provide a space separated '
                             'list. Possible values: none - do not compute; bound_box - compounds with descriptor '
                             'values out of those for training set compounds are outside AD. Default: none')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='show progress on the screen.')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "input": x_fname = v
        if o == "input_format": input_format = v
        if o == "output": out_fname = v
        if o == "models": model_names = v
        if o == "models_dir": model_dir = v
        if o == "model_type": model_type = v
        if o == "applicability_domain": ad = v
        if o == "verbose": verbose = v
    if ad is not None and 'none' in ad:
        ad.remove('none')
        if not ad:
            ad = None

    main_params(x_fname, input_format, out_fname, model_names, model_dir, model_type, ad, verbose)


if __name__ == '__main__':
    main()




import setuptools
from os import path
import spci

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="spci",
    version=spci.__version__,
    author="Pavel Polishchuk",
    author_email="pavel_polishchuk@ukr.net",
    description="SPCI: structural and physicochemical interpretation of QSAR models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DrrDom/spci",
    packages=['spci'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry"
    ],
    python_requires='>=3.6',
    install_requires=['sirms>=1.2.0'],
    extras_require={
        'rdkit': ['rdkit>=2017.09'],
    },
    entry_points={'console_scripts':
                  ['spci = spci.spci:main',
                   'spci_add_relative_frag_size = spci.utils.add_relative_frag_size:main',
                   'spci_calc_atomic_properties_chemaxon = spci.calc_atomic_properties_chemaxon:main',
                   'spci_calc_frag_contrib = spci.calc_frag_contrib:main',
                   'spci_descriptors = spci.descriptors:entry_point',
                   'spci_find_frags_auto_rdkit = spci.find_frags_auto_rdkit:main',
                   'spci_find_frags_rdkit = spci.find_frags_rdkit:main',
                   'spci_find_murcko_rdkit = spci.find_murcko_rdkit:main',
                   'spci_model = spci.model:main',
                   'spci_plot_contributions = spci.plot_contributions:main',
                   'spci_plot_property_distribution = spci.plot_property_distribution:main',
                   'spci_predict = spci.predict:main']},
    include_package_data=True
)



#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import re
import sys
import argparse
from itertools import chain

patt = re.compile('> {1,2}<(.*)>( +\([0-9]+\))?')


def get_fields(molstr):
    # return dict of field names and values
    d = dict()
    i = 0
    while i < len(molstr):
        if patt.fullmatch(molstr[i]):
            d[patt.sub('\\1', molstr[i])] = molstr[i + 1]
            i += 1
        i += 1
    return d


def main_params(in_fname, out_fname, title, field_names, all_fields, skip_value):

    with open(in_fname) as ifs:

        output = []

        molstr = []
        for line in ifs:
            l = line.strip()
            if l != '$$$$':
                molstr.append(l)
            else:
                d = dict()
                fields = get_fields(molstr)
                if title:
                    if molstr[0]:
                        d['Title'] = molstr[0]
                    else:
                        d['Title'] = ''
                if all_fields:
                    d.update(fields)
                else:
                    if field_names:
                        d.update({f: fields[f] for f in field_names if f in fields})
                output.append(d)
                molstr = []

    with open(out_fname, "w") as ofs:
        # get sorted unique field names
        field_names = sorted(list(set(list(chain.from_iterable([list(s.keys()) for s in output])))))
        if 'Title' in field_names:
            field_names.remove('Title')
            field_names.insert(0, 'Title')
        ofs.write("\t".join(field_names) + "\n")
        sys.stdout.write("\t".join(field_names) + "\n")
        for item in output:
            line = [item.get(f, '') for f in field_names]
            if skip_value:
                # replace skipped values (NA) with empty string
                line = [v if v != skip_value else '' for v in line]
                # skip lines having all empty values
                if title:
                    if all(v == '' for v in line[1:]):
                        continue
                else:
                    if all(v == '' for v in line):
                        continue
            ofs.write("\t".join(line) + "\n")
            sys.stdout.write("\t".join(line) + "\n")


def main():
    parser = argparse.ArgumentParser(description='Extract field values from sdf-files.')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=True,
                        help='input sdf file with standardized structures, molecules should have titles')
    parser.add_argument('-o', '--out', metavar='output.txt', required=True,
                        help='output text file with values of specified fields')
    parser.add_argument('-t', '--title', action='store_true', default=False,
                        help='If true then molecules titles will be extracted')
    parser.add_argument('-f', '--field_names', metavar='[field_name_1 field_name_2 ...]',
                        required=False, default=None, nargs='*',
                        help='space separated list of field names for extraction')
    parser.add_argument('-a', '--all_fields', action='store_true', default=False,
                        help='if set (true) all fields will be extracted.')
    parser.add_argument('--skip', default=None,
                        help='specify field value which will be skipped. Usually NA can be set to skip missing values. '
                             'By default no values are skipped.')

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "in": in_fname = v
        if o == "out": out_fname = v
        if o == "title": title = v
        if o == "field_names": field_names = v
        if o == "all_fields": all_fields = v
        if o == "skip": skip_value = v

    main_params(in_fname, out_fname, title, field_names, all_fields, skip_value)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#==============================================================================
# author          : Pavel Polishchuk
# date            : 01-11-2014
# version         : 0.1
# python_version  : 3.2
# copyright       : Pavel Polishchuk 2014
# license         : LGPLv3
#==============================================================================

import argparse
from matplotlib import pyplot as plt
from collections import Counter


def readDataField(sdf_fname, field_names):
    d = dict()
    field_names = [">  <" + el.lower() + ">" for el in field_names]
    with open(sdf_fname) as f:
        line = f.readline()
        while line:
            if line.strip().lower() in field_names:
                field_name = line.strip().lower()[4:-1]
                if field_name in d.keys():
                    d[field_name].extend(f.readline().strip().split(";"))
                else:
                    d[field_name] = f.readline().strip().split(";")
            line = f.readline()

    for field_name in d.keys():
        try:
            d[field_name] = [float(el.replace(",", ".")) for el in d[field_name] if el != ""]
        except ValueError:
            d[field_name] = [el for el in d[field_name] if el != ""]

    return (d)


def plot(data):
    def autolabel(splot, rects):
        # attach some text labels
        for rect in rects:
            height = rect.get_height()
            splot.text(rect.get_x()+rect.get_width()/2., 1.05*height, '%d'%int(height),
                    ha='center', va='bottom')

    fig = plt.figure()
    h = 1  # number of rows on plot
    w = 1  # number of columns on plot
    for k, i in data.items():
        if i:
            # change w and h values
            if len(fig.axes) >= w * h:
                if w == h:
                    w += 1
                else:
                    h += 1
            # resize previous;y added subplots
            for j in range(len(fig.axes)):
                fig.axes[j].change_geometry(h, w, j + 1)
            # add new subplot
            splot = fig.add_subplot(h, w, len(fig.axes) + 1)
            splot.set_xlabel(k)
            if type(i[0]) is float:
                n, bins, patches = splot.hist(i, 50, facecolor='green', alpha=0.5, histtype="bar")
            else:
                c = Counter(i)
                rect = splot.bar(range(len(c)), list(c.values()), width=0.8)
                splot.set_xticks([el + 0.4 for el in range(len(c))])
                splot.set_xticklabels(list(c.keys()))
                autolabel(splot, rect)
    plt.show()


def main():
    parser = argparse.ArgumentParser(description='Plot distribution diagrams for data fields in sdf-file.')
    parser.add_argument('-i', '--in', metavar='input.sdf', required=True,
                        help='input sdf file with standardized structures, molecules should have titles.')
    parser.add_argument('-p', '--properties', metavar='', nargs='*',
                        default=['charge', 'logp', 'acc', 'don', 'atom_polarizability', 'refractivity', 'hb'],
                        help="list of properties (field names) to plot distribution diagrams. Default: 'charge', 'logp', 'acc', 'don', 'atom_polarizability', 'refractivity', 'hb'.")

    args = vars(parser.parse_args())
    for o, v in args.items():
        if o == "in": in_fname = v
        if o == "properties": prop = v

    d = readDataField(in_fname, prop)

    plot(d)


if __name__ == '__main__':
    main()
